@@ -14,11 +14,10 @@ if not sys.platform.startswith("win"):
 import errno
 import os.path
 import ctypes
-from ctypes import WINFUNCTYPE, WinDLL, POINTER, byref, c_int
+from ctypes import WINFUNCTYPE, WinDLL, POINTER, byref, c_int, wintypes, create_unicode_buffer
 from ctypes.wintypes import BOOL, HANDLE, DWORD, LPWSTR, LPCWSTR, LPVOID, POINT, WORD, SMALL_RECT
 
 from xpra.util import envbool
-from xpra.os_util import PYTHON2, PYTHON3
 from xpra.platform.win32 import constants as win32con
 from xpra.platform.win32.common import (
     GetStdHandle,
@@ -47,31 +46,40 @@ REDIRECT_OUTPUT = envbool("XPRA_REDIRECT_OUTPUT", frozen is True and GetConsoleC
 if frozen:
     #cx_freeze paths:
     PATH = os.environ.get("PATH", "").split(os.pathsep)
-    edir = os.path.dirname(sys.executable)
-    def jedir(*paths):
-        return os.path.join(*([edir]+list(paths)))
-    def addsyspath(*paths):
-        v = jedir(*paths)
-        if os.path.exists(v):
-            if v not in sys.path:
-                sys.path.insert(0, v)
-            if os.path.isdir(v) and v not in PATH:
-                PATH.insert(0, v)
-    addsyspath('')
-    addsyspath('lib')
-    os.environ['GI_TYPELIB_PATH'] = jedir('lib', 'girepository-1.0')
+    edir = os.path.abspath(os.path.dirname(sys.executable))
+    libdir = os.path.join(edir, "lib")
+    for d in (libdir, edir):
+        if not os.path.exists(d) or not os.path.isdir(d):
+            continue
+        try:
+            sys.path.remove(d)
+        except:
+            pass
+        sys.path.insert(0, d)
+        try:
+            PATH.remove(d)
+        except:
+            pass
+        PATH.insert(0, d)
+    os.environ['GI_TYPELIB_PATH'] = os.path.join(libdir, "girepository-1.0")
     os.environ["PATH"] = os.pathsep.join(PATH)
+    if not os.environ.get("GTK_THEME") and not os.environ.get("GTK_DEBUG"):
+        for theme in ("Windows-10", "win32"):
+            tdir = os.path.join(edir, "share", "themes", theme)
+            if os.path.exists(tdir):
+                os.environ["GTK_THEME"] = theme
+                break
 
-#don't know why this breaks with Python 3 yet...
-FIX_UNICODE_OUT = envbool("XPRA_FIX_UNICODE_OUT", not REDIRECT_OUTPUT and PYTHON2)
+if REDIRECT_OUTPUT:
+    FIX_UNICODE_OUT = False
+else:
+    #don't know why this breaks with Python 3 yet...
+    FIX_UNICODE_OUT = envbool("XPRA_FIX_UNICODE_OUT", False)
 
 
 def is_wine():
     try:
-        try:
-            import _winreg as winreg
-        except ImportError:
-            import winreg   #@UnresolvedImport @Reimport
+        import winreg   #@UnresolvedImport @Reimport
         hKey = winreg.OpenKey(win32con.HKEY_LOCAL_MACHINE, r"Software\\Wine")   #@UndefinedVariable
         return hKey is not None
     except:
@@ -80,15 +88,35 @@ def is_wine():
     return False
 
 
+def get_csidl_folder(csidl):
+    try:
+        buf = create_unicode_buffer(wintypes.MAX_PATH)
+        shell32 = WinDLL("shell32", use_last_error=True)
+        SHGetFolderPath = shell32.SHGetFolderPathW
+        SHGetFolderPath(0, csidl, None, 0, buf)
+        return buf.value
+    except:
+        return None
+
+def get_common_startmenu_dir():
+    CSIDL_COMMON_STARTMENU = 0x16
+    return get_csidl_folder(CSIDL_COMMON_STARTMENU)
+
+def get_startmenu_dir():
+    CSIDL_STARTMENU = 0xb
+    return get_csidl_folder(CSIDL_STARTMENU)
+
+def get_commonappdata_dir():
+    CSIDL_COMMON_APPDATA = 35
+    return get_csidl_folder(CSIDL_COMMON_APPDATA)
+
+
 prg_name = "Xpra"
 def set_prgname(name):
     global prg_name
     prg_name = name
     try:
         SetConsoleTitleA(name)
-        if PYTHON2:
-            import glib     #@UnresolvedImport
-            glib.set_prgname(name)
     except:
         pass
 
@@ -99,10 +127,6 @@ def not_a_console(handle):
             or GetConsoleMode(handle, byref(DWORD())) == 0)
 
 def fix_unicode_out():
-    if PYTHON3:
-        _unicode = str
-    else:
-        _unicode = unicode  #@UndefinedVariable
     #code found here:
     #http://stackoverflow.com/a/3259271/428751
     import codecs
@@ -194,11 +218,11 @@ def fix_unicode_out():
                 def write(self, text):
                     try:
                         if self._hConsole is None:
-                            if isinstance(text, _unicode):
+                            if isinstance(text, str):
                                 text = text.encode('utf-8')
                             self._stream.write(text)
                         else:
-                            if not isinstance(text, _unicode):
+                            if not isinstance(text, str):
                                 text = str(text).decode('utf-8')
                             remaining = len(text)
                             while remaining:
@@ -324,8 +348,6 @@ def setup_console_event_listener(handler, enable):
         return False
 
 def do_init():
-    if os.environ.get("CRYPTOGRAPHY_OPENSSL_NO_LEGACY") is None:
-        os.environ["CRYPTOGRAPHY_OPENSSL_NO_LEGACY"] = "1"
     if FIX_UNICODE_OUT:
         fix_unicode_out()
     if not REDIRECT_OUTPUT:
@@ -343,6 +365,7 @@ def do_init():
             log_filename = os.path.join(data_dir, (get_prgname() or "Xpra")+".log")
         sys.stdout = open(log_filename, "a")
         sys.stderr = sys.stdout
+        os.environ["XPRA_LOG_FILENAME"] = log_filename
 
 
 MB_ICONEXCLAMATION  = 0x00000030
