@@ -1,6 +1,6 @@
-#!/usr/bin/env python
+#!/usr/bin/env python3
 # This file is part of Xpra.
-# Copyright (C) 2017-2018 Antoine Martin <antoine@xpra.org>
+# Copyright (C) 2017-2020 Antoine Martin <antoine@xpra.org>
 # Xpra is released under the terms of the GNU GPL v2, or, at your option, any
 # later version. See the file COPYING for details.
 
@@ -13,17 +13,14 @@ from ctypes import (
 from ctypes.wintypes import (
     HWND, DWORD, WPARAM, LPARAM, HDC, HMONITOR, HMODULE,
     SHORT, ATOM, RECT, POINT, MAX_PATH, WCHAR, BYTE,
-    )
-from ctypes.wintypes import (
     HANDLE, LPSTR, LPCWSTR, UINT, INT, BOOL, WORD, HGDIOBJ,
     LONG, LPVOID, HBITMAP, LPCSTR, LPWSTR, HWINSTA,
-    HINSTANCE, HMENU, ULONG, HHOOK, MSG,
-    HKL,
+    HINSTANCE, HMENU, ULONG, HHOOK, LPMSG,
+    LCID, HKL,
     )
 #imported from this module but not used here:
 assert GetLastError
 
-LPMSG = POINTER(MSG)
 LPCTSTR = LPCSTR
 LRESULT = c_long
 DEVMODE = c_void_p
@@ -104,6 +101,29 @@ class CIEXYZTRIPLE(Structure):
         ('ciexyzGreen', CIEXYZ),
     ]
 
+class BITMAPINFOHEADER(Structure):
+    _fields_ = [
+        ("biSize",          DWORD),
+        ("biWidth",         LONG),
+        ("biHeight",        LONG),
+        ("biPlanes",        WORD),
+        ("biBitCount",      WORD),
+        ("biCompression",   DWORD),
+        ("biSizeImage",     DWORD),
+        ("biXPelsPerMeter", LONG),
+        ("biYPelsPerMeter", LONG),
+        ("biClrUsed",       DWORD),
+        ("biClrImportant",  DWORD),
+        ]
+PBITMAPINFOHEADER = POINTER(BITMAPINFOHEADER)
+
+class BITMAPINFO(Structure):
+    _fields_ = [
+        ("bmiHeader",       BITMAPINFOHEADER),
+        ("bmiColors",       DWORD),
+        ]
+PBITMAPINFO = POINTER(BITMAPINFO)
+
 class BITMAPV5HEADER(Structure):
     _fields_ = [
         ('bV5Size',             DWORD),
@@ -131,6 +151,7 @@ class BITMAPV5HEADER(Structure):
         ('bV5ProfileSize',      DWORD),
         ('bV5Reserved',         DWORD),
     ]
+PBITMAPV5HEADER = POINTER(BITMAPV5HEADER)
 
 CCHDEVICENAME = 32
 class MONITORINFOEX(Structure):
@@ -212,6 +233,9 @@ GlobalFree.argtypes = [HGLOBAL]
 GlobalLock = kernel32.GlobalLock
 GlobalLock.restype = LPVOID
 GlobalLock.argtypes = [HGLOBAL]
+GlobalSize = kernel32.GlobalSize
+GlobalSize.argtypes = [HGLOBAL]
+GlobalSize.restype = c_size_t
 GlobalUnlock = kernel32.GlobalUnlock
 GlobalUnlock.restype = BOOL
 GlobalUnlock.argtypes = [HGLOBAL]
@@ -224,6 +248,15 @@ MultiByteToWideChar.argtypes =  [UINT, DWORD, LPCSTR, c_int, LPWSTR, c_int]
 WTSGetActiveConsoleSessionId = kernel32.WTSGetActiveConsoleSessionId
 WTSGetActiveConsoleSessionId.restype = DWORD
 WTSGetActiveConsoleSessionId.argtypes = []
+QueryFullProcessImageNameA = kernel32.QueryFullProcessImageNameA
+QueryFullProcessImageNameA.restype = BOOL
+QueryFullProcessImageNameA.argtypes = [HANDLE, DWORD, LPSTR, PDWORD]
+OpenProcess = kernel32.OpenProcess
+OpenProcess.restype = HANDLE
+OpenProcess.argtypes = [DWORD, BOOL, DWORD]
+GetThreadLocale = kernel32.GetThreadLocale
+GetThreadLocale.argtypes = []
+GetThreadLocale.restype = LCID
 
 user32 = WinDLL("user32", use_last_error=True)
 RegisterClassExA = user32.RegisterClassExA
@@ -333,6 +366,7 @@ IsWindowVisible = user32.IsWindowVisible
 GetWindowTextLengthW = user32.GetWindowTextLengthW
 GetWindowTextW = user32.GetWindowTextW
 GetWindowThreadProcessId = user32.GetWindowThreadProcessId
+GetWindowThreadProcessId.argtypes = [HWND, LPDWORD]
 GetWindowThreadProcessId.restype = DWORD
 GetDesktopWindow = user32.GetDesktopWindow
 GetDesktopWindow.restype = HWND
@@ -349,6 +383,8 @@ LoadIconA.restype = HICON
 #LoadIconA.argtypes = [HINSTANCE, LPCSTR]
 RegisterWindowMessageA = user32.RegisterWindowMessageA
 UpdateWindow = user32.UpdateWindow
+UpdateWindow.restype = BOOL
+UpdateWindow.argtypes = [HWND]
 DestroyIcon = user32.DestroyIcon
 DestroyIcon.restype = BOOL
 DestroyIcon.argtypes = [HICON]
@@ -403,7 +439,19 @@ CreateWindowStationA.argtypes = [LPCSTR, DWORD, ACCESS_MASK, LPSECURITY_ATTRIBUT
 CloseWindowStation = user32.CloseWindowStation
 CloseWindowStation.restype = BOOL
 CloseWindowStation.argtypes = [HWINSTA]
+WINSTAENUMPROCA = WINFUNCTYPE(BOOL, LPSTR, LPARAM)
+EnumWindowStationsA = user32.EnumWindowStationsA
+EnumWindowStationsA.restype = BOOL
+EnumWindowStationsA.argtypes = [WINSTAENUMPROCA, LPARAM]
 HDESK = HANDLE
+CreateDesktopA = user32.CreateDesktopA
+CreateDesktopA.restype = HDESK
+DEVMODEA = c_void_p
+CreateDesktopA.argtypes = [LPCSTR, LPCSTR, DEVMODEA, DWORD, ACCESS_MASK, LPSECURITY_ATTRIBUTES]
+CreateDesktopExA = user32.CreateDesktopExA
+CreateDesktopExA.argtypes = [LPCSTR, LPCSTR, DEVMODEA, DWORD, ACCESS_MASK, LPSECURITY_ATTRIBUTES, ULONG, LPVOID]
+CreateDesktopExA = user32.CreateDesktopExA
+CreateDesktopExA.restype = HDESK
 OpenDesktopW = user32.OpenDesktopW
 OpenDesktopW.restype = HDESK
 OpenDesktopW.argtypes = [LPWSTR, DWORD, BOOL, ACCESS_MASK]
@@ -413,6 +461,10 @@ CloseDesktop.argtypes = [HDESK]
 OpenInputDesktop = user32.OpenInputDesktop
 OpenInputDesktop.restype = HDESK
 OpenInputDesktop.argtypes = [DWORD, BOOL, ACCESS_MASK]
+DESKTOPENUMPROCA = WINFUNCTYPE(BOOL, LPCSTR, LPARAM)
+EnumDesktopsA = user32.EnumDesktopsA
+EnumDesktopsA.restype = BOOL
+EnumDesktopsA.argtypes = [HWINSTA, DESKTOPENUMPROCA, LPARAM]
 GetUserObjectInformationA = user32.GetUserObjectInformationA
 GetUserObjectInformationA.restype = BOOL
 GetUserObjectInformationA.argtypes = [HANDLE, INT, LPVOID, DWORD, LPDWORD]
@@ -444,12 +496,6 @@ SetClipboardData.argtypes = [UINT, HANDLE]
 RegisterClipboardFormatA = user32.RegisterClipboardFormatA
 RegisterClipboardFormatA.restype = UINT
 RegisterClipboardFormatA.argtypes = [LPCSTR]
-GetClipboardFormatNameA = user32.GetClipboardFormatNameA
-GetClipboardFormatNameA.restype = c_int
-GetClipboardFormatNameA.argtypes = [UINT, LPCTSTR, c_int]
-RegisterClipboardFormatA = user32.RegisterClipboardFormatA
-RegisterClipboardFormatA.restype = UINT
-RegisterClipboardFormatA.argtypes = [LPCSTR]
 GetClipboardData = user32.GetClipboardData
 GetClipboardData.restype = HANDLE
 GetClipboardData.argtypes = [UINT]
@@ -489,6 +535,19 @@ EnumClipboardFormats.argtypes = [UINT]
 CountClipboardFormats = user32.CountClipboardFormats
 CountClipboardFormats.restype = c_int
 CountClipboardFormats.argtypes = []
+GetWindowTextA = user32.GetWindowTextA
+GetWindowTextA.restype = c_int
+GetWindowTextA.argtypes = [HWND, LPSTR, c_int]
+GetWindowTextW = user32.GetWindowTextW
+GetWindowTextW.restype = c_int
+GetWindowTextW.argtypes = [HWND, LPWSTR, c_int]
+GetWindowTextLengthA = user32.GetWindowTextLengthA
+GetWindowTextLengthA.restype = c_int
+GetWindowTextLengthA.argtypes = [HWND]
+GetWindowTextLengthW = user32.GetWindowTextLengthW
+GetWindowTextLengthW.restype = c_int
+GetWindowTextLengthW.argtypes = [HWND]
+
 
 class PROCESS_INFORMATION(Structure):
     _fields_ = (
@@ -524,6 +583,8 @@ advapi32 = WinDLL("advapi32")
 CreateProcessAsUserA = advapi32.CreateProcessAsUserA
 CreateProcessAsUserA.restype = BOOL
 CreateProcessAsUserA.argtypes = [HANDLE, LPCSTR, LPSTR, LPSECURITY_ATTRIBUTES, LPSECURITY_ATTRIBUTES, BOOL, DWORD, LPVOID, LPCSTR, LPCSTR, PSTARTUPINFOA, PPROCESS_INFORMATION]
+GetUserNameA = advapi32.GetUserNameA
+GetUserNameA.restype = BOOL
 
 gdi32 = WinDLL("gdi32", use_last_error=True)
 CreateCompatibleDC = gdi32.CreateCompatibleDC
@@ -556,6 +617,9 @@ DeleteDC = gdi32.DeleteDC
 CreateDIBSection = gdi32.CreateDIBSection
 CreateDIBSection.restype = HBITMAP
 CreateDIBSection.argtypes = [HANDLE, POINTER(BITMAPV5HEADER), UINT, POINTER(c_void_p), HANDLE, DWORD]
+CreateDIBitmap = gdi32.CreateDIBitmap
+CreateDIBitmap.restype = HBITMAP
+CreateDIBitmap.argtypes = [HDC, PBITMAPINFOHEADER, DWORD, c_void_p, PBITMAPINFO, UINT]
 DeleteObject = gdi32.DeleteObject
 DeleteObject.argtypes = [HGDIOBJ]
 DeleteObject.restype = BOOL
