@@ -61,8 +61,11 @@ class Keyboard(KeyboardBase):
         KEY_TRANSLATIONS[("dead_tilde", 65107,  50)]    = "asciitilde"
         KEY_TRANSLATIONS[("dead_grave", 65104,  55)]    = "grave"
 
+    def __repr__(self):
+        return "win32.Keyboard"
+
     def set_modifier_mappings(self, mappings):
-        KeyboardBase.set_modifier_mappings(self, mappings)
+        super().set_modifier_mappings(mappings)
         self.num_lock_modifier = self.modifier_keys.get("Num_Lock")
         log("set_modifier_mappings found 'Num_Lock' with modifier value: %s", self.num_lock_modifier)
         for x in ("ISO_Level3_Shift", "Mode_switch"):
@@ -74,7 +77,7 @@ class Keyboard(KeyboardBase):
 
     def mask_to_names(self, mask):
         """ Patch NUMLOCK and AltGr """
-        names = KeyboardBase.mask_to_names(self, mask)
+        names = super().mask_to_names(mask)
         if EMULATE_ALTGR:
             rmenu = GetKeyState(win32con.VK_RMENU)
             #log("GetKeyState(VK_RMENU)=%s", rmenu)
@@ -88,7 +91,7 @@ class Keyboard(KeyboardBase):
                 elif not numlock and self.num_lock_modifier in names:
                     names.remove(self.num_lock_modifier)
                 log("mask_to_names(%s) GetKeyState(VK_NUMLOCK)=%s, names=%s", mask, numlock, names)
-            except:
+            except Exception:
                 pass
         else:
             log("mask_to_names(%s)=%s", mask, names)
@@ -117,6 +120,19 @@ class Keyboard(KeyboardBase):
         """
         return  {}, [], ["lock"]
 
+
+    def get_all_x11_layouts(self):
+        x11_layouts = {}
+        for win32_layout in WIN32_LAYOUTS.values():
+            #("ARA", "Saudi Arabia",   "Arabic",                   1356,   "ar", []),
+            x11_layout = win32_layout[4]
+            if x11_layout in x11_layouts:
+                continue
+            name = win32_layout[2]
+            x11_layouts[x11_layout] = name
+        return x11_layouts
+
+
     def get_layout_spec(self):
         KMASKS = (0xffffffff, 0xffff, 0x3ff)
         layout = None
@@ -249,9 +265,8 @@ class Keyboard(KeyboardBase):
                     if rmenu in (0, 1):
                         self.delayed_event = (send_key_action_cb, wid, key_event)
                         #needed for altgr emulation timeouts:
-                        from xpra.gtk_common.gobject_compat import import_glib
-                        glib = import_glib()
-                        glib.timeout_add(EMULATE_ALTGR_CONTROL_KEY_DELAY, self.send_delayed_key)
+                        from gi.repository import GLib
+                        GLib.timeout_add(EMULATE_ALTGR_CONTROL_KEY_DELAY, self.send_delayed_key)
                     return
                 if not key_event.pressed and rmenu not in (0, 1):
                     #unpressed: could just skip it?
@@ -271,7 +286,7 @@ class Keyboard(KeyboardBase):
                 key_event.keycode = -1
                 self.AltGr_modifiers(key_event.modifiers)
         self.send_delayed_key()
-        KeyboardBase.process_key_event(self, send_key_action_cb, wid, key_event)
+        super().process_key_event(send_key_action_cb, wid, key_event)
 
     def send_delayed_key(self):
         #timeout: this must be a real one, send it now
@@ -282,4 +297,4 @@ class Keyboard(KeyboardBase):
             rmenu = GetKeyState(win32con.VK_RMENU)
             log("send_delayed_key() GetKeyState(VK_RMENU)=%s", rmenu)
             if rmenu in (0, 1):
-                KeyboardBase.process_key_event(self, *dk)
+                super().process_key_event(*dk)
