@@ -1,10 +1,11 @@
 # This file is part of Xpra.
 # Copyright (C) 2010 Nathaniel Smith <njs@pobox.com>
-# Copyright (C) 2011-2019 Antoine Martin <antoine@xpra.org>
+# Copyright (C) 2011-2021 Antoine Martin <antoine@xpra.org>
 # Xpra is released under the terms of the GNU GPL v2, or, at your option, any
 # later version. See the file COPYING for details.
 
-from xpra.util import nonl
+import os
+
 from xpra.platform.keyboard_base import KeyboardBase
 from xpra.keyboard.mask import MODIFIER_MAP
 from xpra.keyboard.layouts import xkbmap_query_tostring
@@ -28,6 +29,8 @@ class Keyboard(KeyboardBase):
                 log.error("Error: failed to load posix keyboard bindings")
                 log.error(" %s", str(e) or type(e))
 
+    def __repr__(self):
+        return "xposix.Keyboard"
 
     def get_keymap_modifiers(self):
         if self.keymap_modifiers is None:
@@ -36,7 +39,10 @@ class Keyboard(KeyboardBase):
 
     def do_get_keymap_modifiers(self):
         if not self.keyboard_bindings:
-            log.warn("keyboard bindings are not available, expect keyboard mapping problems")
+            log.warn("keyboard bindings are not available")
+            log.warn(" expect keyboard mapping problems")
+            if is_Wayland():
+                log.warn(" (incomplete wayland support)")
             return {}, [], []
         try:
             from xpra.gtk_common.error import xsync
@@ -76,12 +82,8 @@ class Keyboard(KeyboardBase):
         # System Locale: LANG=en_GB.UTF-8
         # VC Keymap: gb
         # X11 Layout: gb
-        from subprocess import check_output
-        try:
-            out = check_output(["localectl", "status"])
-        except Exception:
-            log("get_locale_status() 'localectl status' failed", exc_info=True)
-            return {}
+        from subprocess import getoutput
+        out = getoutput("localectl status")
         if not out:
             return {}
         locale = {}
@@ -105,17 +107,15 @@ class Keyboard(KeyboardBase):
             return None, None, query_struct
         query_struct = self.keyboard_bindings.getXkbProperties()
         _query = xkbmap_query_tostring(query_struct)
-        log("get_keymap_spec() Xkb query tostring(%s)=%r", query_struct, _query)
+        log("get_keymap_spec() Xkb query tostring(%s)=%s", query_struct, _query)
         #we no longer support servers via xkbmap_print:
         xkbmap_print = ""
-        log("get_keymap_spec()=(%s, %s, %s)", nonl(xkbmap_print), nonl(_query), nonl(query_struct))
+        log("get_keymap_spec()=(%r, %r, %r)", xkbmap_print, _query, query_struct)
         return xkbmap_print, _query, query_struct
 
 
     def get_xkb_rules_names_property(self):
         #parses the "_XKB_RULES_NAMES" X11 property
-        #FIXME: a bit ugly to call gtk here...
-        #but otherwise we have to call XGetWindowProperty and deal with X11 errors..
         if not is_X11():
             return ""
         xkb_rules_names = ""
@@ -124,19 +124,57 @@ class Keyboard(KeyboardBase):
         log("get_xkb_rules_names_property() _XKB_RULES_NAMES=%s", prop)
         #ie: 'evdev\x00pc104\x00gb,us\x00,\x00\x00'
         if prop:
-            xkb_rules_names = bytestostr(prop).split("\0")
+            xkb_rules_names = prop.split("\0")
             #ie: ['evdev', 'pc104', 'gb,us', ',', '', '']
         log("get_xkb_rules_names_property()=%s", xkb_rules_names)
         return xkb_rules_names
 
+
+    def get_all_x11_layouts(self):
+        repository = "/usr/share/X11/xkb/rules/base.xml"
+        if os.path.exists(repository):
+            try:
+                import lxml.etree  #pylint: disable=import-outside-toplevel
+            except ImportError:
+                log("cannot parse xml", exc_info=True)
+            else:
+                with open(repository) as f:
+                    tree = lxml.etree.parse(f)  #pylint: disable=c-extension-no-member
+                x11_layouts = {}
+                for layout in tree.xpath("//layout"):
+                    layout = layout.xpath("./configItem/name")[0].text
+                    x11_layouts[layout] = layout
+                    #for variant in layout.xpath("./variantList/variant/configItem/name"):
+                    #    variant_name = variant.text
+                return x11_layouts
+        from subprocess import Popen, PIPE  #pylint: disable=import-outside-toplevel
+        try:
+            proc = Popen(["localectl", "list-x11-keymap-layouts"], stdout=PIPE, stderr=PIPE)
+            out = proc.communicate()[0]
+            log("get_all_x11_layouts() proc=%s", proc)
+            log("get_all_x11_layouts() returncode=%s", proc.returncode)
+            if proc.wait()==0:
+                x11_layouts = {}
+                for line in out.splitlines():
+                    layout = line.decode().split("/")[-1]
+                    if layout:
+                        x11_layouts[layout] = layout
+                return x11_layouts
+        except OSError:
+            log("get_all_x11_layouts()", exc_info=True)
+        return {"us" : "English"}
+
+
     def get_layout_spec(self):
         layout = ""
         layouts = []
+        variant = ""
         options = ""
         v = None
         if self.keyboard_bindings:
             props = self.keyboard_bindings.getXkbProperties()
             v = props.get("layout")
+            variant = props.get("variant", "")
             options = props.get("options", "")
         else:
             locale = self.get_locale_status()
@@ -153,9 +191,9 @@ class Keyboard(KeyboardBase):
         def s(v):
             try:
                 return v.encode("latin1")
-            except:
+            except Exception:
                 return str(v)
-        return s(layout), [s(x) for x in layouts], "", None, options
+        return s(layout), [s(x) for x in layouts], variant, None, options
 
 
     def get_keyboard_repeat(self):
@@ -182,6 +220,6 @@ class Keyboard(KeyboardBase):
         self.keymap_modifiers = None
         try:
             dn = "%s %s" % (type(display).__name__, display.get_name())
-        except:
+        except Exception:
             dn = str(display)
         log("update_modifier_map(%s, %s) modifier_map=%s", dn, xkbmap_mod_meanings, self.modifier_map)
