@@ -1,34 +1,39 @@
 # -*- coding: utf-8 -*-
 # This file is part of Xpra.
-# Copyright (C) 2010-2019 Antoine Martin <antoine@xpra.org>
+# Copyright (C) 2010-2021 Antoine Martin <antoine@xpra.org>
 # Copyright (C) 2008 Nathaniel Smith <njs@pobox.com>
 # Xpra is released under the terms of the GNU GPL v2, or, at your option, any
 # later version. See the file COPYING for details.
 
+import shlex
 import os.path
 
 from xpra.platform.features import COMMAND_SIGNALS
 from xpra.child_reaper import getChildReaper, reaper_cleanup
-from xpra.os_util import monotonic_time, bytestostr, OSX, WIN32, POSIX
-from xpra.util import envint, csv
-from xpra.scripts.parsing import parse_env
+from xpra.os_util import (
+    monotonic_time, bytestostr,
+    OSX, WIN32, POSIX,
+    )
+from xpra.util import envint, csv, ellipsizer
 from xpra.make_thread import start_thread
+from xpra.scripts.parsing import parse_env, get_subcommands
 from xpra.server.server_util import source_env
+from xpra.server.menu_provider import get_menu_provider
 from xpra.server import EXITING_CODE
 from xpra.server.mixins.stub_server_mixin import StubServerMixin
 from xpra.log import Logger
 
 log = Logger("exec")
+httplog = Logger("http")
 
 TERMINATE_DELAY = envint("XPRA_TERMINATE_DELAY", 1000)/1000.0
-MENU_RELOAD_DELAY = envint("XPRA_MENU_RELOAD_DELAY", 5)
 
 
-"""
-Mixin for servers that can handle file transfers and forwarded printers.
-Printer forwarding is only supported on Posix servers with the cups backend script.
-"""
 class ChildCommandServer(StubServerMixin):
+    """
+    Mixin for servers that can handle file transfers and forwarded printers.
+    Printer forwarding is only supported on Posix servers with the cups backend script.
+    """
 
     def __init__(self):
         self.child_display = None
@@ -52,14 +57,12 @@ class ChildCommandServer(StubServerMixin):
         self.children_started = []
         self.child_reaper = None
         self.reaper_exit = self.reaper_exit_check
-        self.watch_manager = None
-        self.watch_notifier = None
-        self.xdg_menu_reload_timer = None
         #does not belong here...
         if not hasattr(self, "_upgrading"):
             self._upgrading = False
         if not hasattr(self, "session_name"):
             self.session_name = ""
+        self.menu_provider = None
 
     def init(self, opts):
         self.exit_with_children = opts.exit_with_children
@@ -74,11 +77,14 @@ class ChildCommandServer(StubServerMixin):
         self.start_on_last_client_exit   = opts.start_on_last_client_exit
         self.start_child_on_last_client_exit = opts.start_child_on_last_client_exit
         if opts.exec_wrapper:
-            import shlex
             self.exec_wrapper = shlex.split(opts.exec_wrapper)
         self.child_reaper = getChildReaper()
         self.source_env = source_env(opts.source_start)
         self.start_env = parse_env(opts.start_env)
+        if self.start_new_commands:
+            #may already have been initialized by servercore:
+            self.menu_provider = self.menu_provider or get_menu_provider()
+            self.menu_provider.on_reload.append(self.send_updated_menu)
 
     def threaded_setup(self):
         self.exec_start_commands()
@@ -86,51 +92,6 @@ class ChildCommandServer(StubServerMixin):
             self.child_reaper.set_quit_callback(self.reaper_exit)
             self.child_reaper.check()
         self.idle_add(set_reaper_callback)
-        if POSIX and not OSX and self.start_new_commands:
-            try:
-                self.setup_menu_watcher()
-            except Exception as e:
-                log("threaded_setup()", exc_info=True)
-                log.error("Error setting up menu watcher:")
-                log.error(" %s", e)
-            from xpra.platform.xposix.xdg_helper import load_xdg_menu_data
-            load_xdg_menu_data()
-
-    def setup_menu_watcher(self):
-        try:
-            import pyinotify
-        except ImportError as e:
-            log.warn("Warning: cannot watch for application menu changes without pyinotify:")
-            log.warn(" %s", e)
-            return
-        self.watch_manager = pyinotify.WatchManager()
-        def menu_data_updated(create, pathname):
-            log("menu_data_updated(%s, %s)", create, pathname)
-            self.schedule_xdg_menu_reload()
-        class EventHandler(pyinotify.ProcessEvent):
-            def process_IN_CREATE(self, event):
-                menu_data_updated(True, event.pathname)
-            def process_IN_DELETE(self, event):
-                menu_data_updated(False, event.pathname)
-        mask = pyinotify.IN_DELETE | pyinotify.IN_CREATE  #@UndefinedVariable pylint: disable=no-member
-        handler = EventHandler()
-        self.watch_notifier = pyinotify.ThreadedNotifier(self.watch_manager, handler)
-        self.watch_notifier.setDaemon(True)
-        data_dirs = os.environ.get("XDG_DATA_DIRS", "/usr/share:/usr/local/share").split(":")
-        watched = []
-        for data_dir in data_dirs:
-            menu_dir = os.path.join(data_dir, "applications")
-            if not os.path.exists(menu_dir):
-                continue
-            wdd = self.watch_manager.add_watch(menu_dir, mask)
-            watched.append(menu_dir)
-            log("watch_notifier=%s, watch=%s", self.watch_notifier, wdd)
-        self.watch_notifier.start()
-        if watched:
-            log.info("watching for applications menu changes in:")
-            for wd in watched:
-                log.info(" '%s'", wd)
-
 
     def cleanup(self):
         if self.terminate_children and self._upgrading!=EXITING_CODE:
@@ -138,25 +99,12 @@ class ChildCommandServer(StubServerMixin):
         def noop():
             pass
         self.reaper_exit = noop
+        if self.menu_provider:
+            self.menu_provider.cleanup()
         reaper_cleanup()
-        xmrt = self.xdg_menu_reload_timer
-        if xmrt:
-            self.xdg_menu_reload_timer = None
-            self.source_remove(xmrt)
-        wn = self.watch_notifier
-        if wn:
-            self.watch_notifier = None
-            wn.stop()
-        watch_manager = self.watch_manager
-        if watch_manager:
-            self.watch_manager = None
-            try:
-                watch_manager.close()
-            except (OSError, IOError):
-                log("error closing watch manager %s", watch_manager, exc_info=True)
 
 
-    def get_server_features(self, _source):
+    def get_server_features(self, _source) -> dict:
         return {
             "start-new-commands"        : self.start_new_commands,
             "exit-with-children"        : self.exit_with_children,
@@ -167,57 +115,46 @@ class ChildCommandServer(StubServerMixin):
 
 
     def _get_xdg_menu_data(self, force_reload=False):
-        if not self.start_new_commands or not POSIX or OSX:
+        if not self.start_new_commands:
             return None
-        from xpra.platform.xposix.xdg_helper import load_xdg_menu_data
-        return load_xdg_menu_data(force_reload)
+        assert self.menu_provider
+        return self.menu_provider.get_menu_data(force_reload)
 
-    def get_caps(self, source):
+
+    def get_caps(self, source) -> dict:
         caps = {}
         if not source:
             return caps
         #don't assume we have a real ClientConnection object:
-        if getattr(source, "wants_features", False):
+        if getattr(source, "wants_features", False) and getattr(source, "ui_client", False):
             caps["xdg-menu"] = {}
             if not source.xdg_menu_update:
-                #we have to send it now:
-                xdg_menu = self._get_xdg_menu_data()
-                if xdg_menu:
-                    l = len(str(xdg_menu))
-                    #arbitrary: don't use more than half
-                    #of the maximum size of the hello packet:
-                    if l>2*1024*1024:
-                        from xpra.platform.xposix.xdg_helper import remove_icons
-                        xdg_menu = remove_icons(xdg_menu)
-                        log.info("removed icons to reduce the size of the xdg menu data")
-                        log.info("size reduced from %i to %i", l, len(str(xdg_menu)))
-                    caps["xdg-menu"] = xdg_menu
+                log.warn("Warning: outdated client does not support xdg-menu update")
+            caps["subcommands"] = get_subcommands()
         return caps
 
     def send_initial_data(self, ss, caps, send_ui, share_count):
-        xdg_menu = self._get_xdg_menu_data()
         if ss.xdg_menu_update:
-            ss.send_setting_change("xdg-menu", xdg_menu or {})
+            #this method may block if the menus are still being loaded,
+            #so do it in a throw-away thread:
+            start_thread(self.send_xdg_menu_data, "send-xdg-menu-data", True, (ss,))
 
-    def schedule_xdg_menu_reload(self):
-        xmrt = self.xdg_menu_reload_timer
-        if xmrt:
-            self.source_remove(xmrt)
-        self.xdg_menu_reload_timer = self.timeout_add(MENU_RELOAD_DELAY*1000, self.xdg_menu_reload)
+    def send_xdg_menu_data(self, ss):
+        if ss.is_closed():
+            return
+        xdg_menu = self._get_xdg_menu_data() or {}
+        log("%i entries sent in initial data", len(xdg_menu))
+        ss.send_setting_change("xdg-menu", xdg_menu)
 
-    def xdg_menu_reload(self):
-        self.xdg_menu_reload_timer = None
-        log("xdg_menu_reload()")
-        start_thread(self.do_xdg_menu_reload, "reload-xdg-menu-data", True)
-
-    def do_xdg_menu_reload(self):
-        xdg_menu = self._get_xdg_menu_data(True)
+    def send_updated_menu(self, xdg_menu):
+        log("send_updated_menu(%s)", ellipsizer(xdg_menu))
         for source in tuple(self._server_sources.values()):
             #some sources don't have this attribute (ie: RFBSource does not):
             if getattr(source, "xdg_menu_update", False):
                 source.send_setting_change("xdg-menu", xdg_menu or {})
 
-    def get_info(self, _proto):
+
+    def get_info(self, _proto) -> dict:
         info = {
             "start"                     : self.start_commands,
             "start-child"               : self.start_child_commands,
@@ -229,6 +166,12 @@ class ChildCommandServer(StubServerMixin):
             "start-after-connect-done"  : self.start_after_connect_done,
             "start-new"                 : self.start_new_commands,
             }
+        mp = self.menu_provider
+        if mp:
+            info.update({
+            "start-menu"                : mp.get_menu_data(remove_icons=True, wait=False) or {},
+            "start-desktop-menu"        : mp.get_desktop_sessions(remove_icons=True) or {},
+            })
         for i,procinfo in enumerate(self.children_started):
             info[i] = procinfo.get_info()
         return {"commands": info}
@@ -240,19 +183,16 @@ class ChildCommandServer(StubServerMixin):
 
     def get_child_env(self):
         #subclasses may add more items (ie: fakexinerama)
-        env = os.environ.copy()
+        env = super().get_child_env()
         env.update(self.source_env)
         env.update(self.start_env)
         if self.child_display:
             env["DISPLAY"] = self.child_display
         return env
 
-
-    def get_full_child_command(self, cmd, use_wrapper=True):
+    def get_full_child_command(self, cmd, use_wrapper=True) -> list:
         #make sure we have it as a list:
-        if not isinstance(cmd, (list, tuple)):
-            import shlex
-            cmd = shlex.split(str(cmd))
+        cmd = super().get_full_child_command(cmd, use_wrapper)
         if not use_wrapper or not self.exec_wrapper:
             return cmd
         return self.exec_wrapper + cmd
@@ -290,17 +230,18 @@ class ChildCommandServer(StubServerMixin):
             self.idle_add(self.guess_session_name, procs)
 
     def start_command(self, name, child_cmd, ignore=False, callback=None, use_wrapper=True, shell=False, **kwargs):
-        log("start_command%s exec_wrapper=%s",
-            (name, child_cmd, ignore, callback, use_wrapper, shell, kwargs), self.exec_wrapper)
         from subprocess import Popen
         env = self.get_child_env()
+        log("start_command%s exec_wrapper=%s, exec_cwd=%s",
+            (name, child_cmd, ignore, callback, use_wrapper, shell, kwargs), self.exec_wrapper, self.exec_cwd)
+        real_cmd = []
         try:
             real_cmd = self.get_full_child_command(child_cmd, use_wrapper)
             log("full child command(%s, %s)=%s", child_cmd, use_wrapper, real_cmd)
-            proc = Popen(real_cmd, env=env, shell=shell, cwd=self.exec_cwd, close_fds=True, **kwargs)
+            proc = Popen(real_cmd, env=env, shell=shell, cwd=self.exec_cwd, **kwargs)
             procinfo = self.add_process(proc, name, real_cmd, ignore=ignore, callback=callback)
             log("pid(%s)=%s", real_cmd, proc.pid)
-            log.info("started command '%s' with pid %s", " ".join(real_cmd), proc.pid)
+            log.info("started command '%s' with pid %s", " ".join(bytestostr(x) for x in real_cmd), proc.pid)
             if not ignore:
                 self.children_count += 1
             self.children_started.append(procinfo)
@@ -308,6 +249,9 @@ class ChildCommandServer(StubServerMixin):
         except (OSError, ValueError) as e:
             log("start_command%s", (name, child_cmd, ignore, callback, use_wrapper, shell, kwargs), exc_info=True)
             log.error("Error spawning child '%s':" % (child_cmd, ))
+            if len(real_cmd)>1:
+                log.error(" using command:")
+                log.error(" %s", " ".join("'%s'" % x for x in real_cmd))
             log.error(" %s" % (e,))
             return None
 
@@ -315,7 +259,7 @@ class ChildCommandServer(StubServerMixin):
     def add_process(self, process, name, command, ignore=False, callback=None):
         return self.child_reaper.add_process(process, name, command, ignore, callback=callback)
 
-    def is_child_alive(self, proc):
+    def is_child_alive(self, proc) -> bool:
         return proc is not None and proc.poll() is None
 
     def reaper_exit_check(self):
