@@ -1,10 +1,10 @@
 # -*- coding: utf-8 -*-
 # This file is part of Xpra.
-# Copyright (C) 2010-2019 Antoine Martin <antoine@xpra.org>
+# Copyright (C) 2010-2020 Antoine Martin <antoine@xpra.org>
 # Xpra is released under the terms of the GNU GPL v2, or, at your option, any
 # later version. See the file COPYING for details.
 
-from xpra.util import envint
+from xpra.util import envint, typedict
 from xpra.server.source.stub_source_mixin import StubSourceMixin
 from xpra.log import Logger
 
@@ -16,6 +16,13 @@ DEFAULT_AV_SYNC_DELAY = envint("XPRA_DEFAULT_AV_SYNC_DELAY", 150)
 
 class AVSyncMixin(StubSourceMixin):
 
+    @classmethod
+    def is_needed(cls, caps : typedict) -> bool:
+        if not (caps.boolget("sound.send") or caps.boolget("sound.receive")):
+            #no audio!
+            return False
+        return caps.boolget("av-sync") and caps.boolget("windows")
+
     def __init__(self):
         self.av_sync = False
 
@@ -32,7 +39,7 @@ class AVSyncMixin(StubSourceMixin):
         self.av_sync_delta = AV_SYNC_DELTA
 
 
-    def get_info(self):
+    def get_info(self) -> dict:
         return {
             "av-sync" : {
                 ""          : self.av_sync,
@@ -43,7 +50,7 @@ class AVSyncMixin(StubSourceMixin):
                 },
             }
 
-    def parse_client_caps(self, c):
+    def parse_client_caps(self, c : typedict):
         av_sync = c.boolget("av-sync")
         self.av_sync_enabled = self.av_sync and av_sync
         self.set_av_sync_delay(int(self.av_sync_enabled) * c.intget("av-sync.delay.default", DEFAULT_AV_SYNC_DELAY))
