@@ -1,13 +1,12 @@
 # -*- coding: utf-8 -*-
 # This file is part of Xpra.
-# Copyright (C) 2010-2019 Antoine Martin <antoine@xpra.org>
+# Copyright (C) 2010-2020 Antoine Martin <antoine@xpra.org>
 # Xpra is released under the terms of the GNU GPL v2, or, at your option, any
 # later version. See the file COPYING for details.
 
 from xpra.os_util import POSIX, OSX, bytestostr
-from xpra.util import envint, csv
+from xpra.util import envint, csv, typedict
 from xpra.server.source.stub_source_mixin import StubSourceMixin
-from xpra.codecs.pillow.decoder import HEADERS, open_only
 from xpra.log import Logger
 
 log = Logger("webcam")
@@ -17,6 +16,7 @@ MAX_WEBCAM_DEVICES = envint("XPRA_MAX_WEBCAM_DEVICES", 1)
 
 def valid_encodings(args):
     #ensure that the encodings specified can be validated using HEADERS
+    from xpra.codecs.pillow.decoder import HEADERS
     encodings = []
     for x in args:
         x = bytestostr(x)
@@ -27,10 +27,23 @@ def valid_encodings(args):
     return encodings
 
 
-"""
-Handle webcam forwarding.
-"""
 class WebcamMixin(StubSourceMixin):
+    """
+    Handle webcam forwarding.
+    """
+
+    @classmethod
+    def is_needed(cls, caps : typedict) -> bool:
+        #the 'webcam' capability was only added in v4,
+        #so we have to enable the mixin by default:
+        if not caps.boolget("webcam", True):
+            return False
+        try:
+            from xpra.codecs.pillow.decoder import HEADERS
+            assert HEADERS
+        except ImportError:
+            return False
+        return True
 
     def __init__(self):
         self.webcam_enabled = False
@@ -52,7 +65,7 @@ class WebcamMixin(StubSourceMixin):
         self.stop_all_virtual_webcams()
 
 
-    def get_info(self):
+    def get_info(self) -> dict:
         return {
             "webcam" : {
                 "encodings"         : self.webcam_encodings,
@@ -61,7 +74,7 @@ class WebcamMixin(StubSourceMixin):
             }
 
 
-    def get_device_options(self, device_id):
+    def get_device_options(self, device_id : int):  #pylint: disable=unused-argument
         if not POSIX or OSX or not self.webcam_enabled:
             return {}
         if self.webcam_device:
@@ -137,15 +150,11 @@ class WebcamMixin(StubSourceMixin):
             self.stop_virtual_webcam(device_id)
 
     def stop_virtual_webcam(self, device_id, message=""):
-        webcam = self.webcam_forwarding_devices.get(device_id)
+        webcam = self.webcam_forwarding_devices.pop(device_id, None)
         log("stop_virtual_webcam(%s, %s) webcam=%s", device_id, message, webcam)
         if not webcam:
             log.warn("Warning: cannot stop webcam device %s: no such context!", device_id)
             return
-        try:
-            del self.webcam_forwarding_devices[device_id]
-        except KeyError:
-            pass
         try:
             webcam.clean()
         except Exception as e:
@@ -162,6 +171,7 @@ class WebcamMixin(StubSourceMixin):
             self.send_webcam_stop(device_id, "not started")
             return False
         try:
+            from xpra.codecs.pillow.decoder import open_only
             assert encoding in self.webcam_encodings, "invalid encoding specified: %s (must be one of %s)" % (encoding, self.webcam_encodings)
             rgb_pixel_format = "BGRX"       #BGRX
             img = open_only(data, (encoding,))
