@@ -127,29 +127,33 @@ class SSHServer(paramiko.ServerInterface):
 
     def check_channel_exec_request(self, channel, command):
         def chan_send(send_fn, data, timeout=5):
+            if not data:
+                return
+            size = len(data)
             start = monotonic_time()
             while data and monotonic_time()-start<timeout:
                 sent = send_fn(data)
+                log("chan_send: sent %i bytes out of %i using %s", sent, size, send_fn)
                 if not sent:
                     break
                 data = data[sent:]
+            if data:
+                raise Exception("failed to send all the data using %s" % send_fn)
         #TODO: close channel after use? when?
         log("check_channel_exec_request(%s, %s)", channel, command)
         try:
             cmd = shlex.split(command.decode("utf8"))
         except UnicodeDecodeError:
             cmd = shlex.split(bytestostr(command))
-        if cmd[0]=="command" and len(cmd)==1:
-            channel.send_exit_status(0)
-            return True
-        if cmd[0] in ("type", "which", "command") and len(cmd)==2:
-            xpra_cmd = cmd[-1]   #ie: $XDG_RUNTIME_DIR/xpra/run-xpra or "xpra"
+        log("check_channel_exec_request: cmd=%s", cmd)
+        if cmd[0] in ("type", "which") and len(cmd)==2:
+            xpra_cmd = cmd[1]   #ie: $XDG_RUNTIME_DIR/xpra/run-xpra or "xpra"
             if not POSIX:
                 assert WIN32
                 #we can't execute "type" or "which" on win32,
                 #so we just answer as best we can
                 #and only accept "xpra" as argument:
-                if xpra_cmd.strip("\"")=="xpra":
+                if xpra_cmd.strip('"').strip("'")=="xpra":
                     chan_send(channel.send, "xpra is xpra")
                     channel.send_exit_status(0)
                 else:
@@ -160,13 +164,15 @@ class SSHServer(paramiko.ServerInterface):
             #but we need to expand the file argument:
             cmd[-1] = osexpand(xpra_cmd)
             try:
-                proc = Popen(cmd, stdin=None, stdout=PIPE, stderr=PIPE, close_fds=not WIN32, shell=False)
+                proc = Popen(cmd, stdout=PIPE, stderr=PIPE, close_fds=not WIN32)
                 out, err = proc.communicate()
             except Exception as e:
                 log("check_channel_exec_request(%s, %s)", channel, command, exc_info=True)
                 chan_send(channel.send_stderr, "failed to execute command: %s" % e)
                 channel.send_exit_status(1)
             else:
+                log("check_channel_exec_request: out(%s)=%s", cmd, out)
+                log("check_channel_exec_request: err(%s)=%s", cmd, err)
                 chan_send(channel.send, out)
                 chan_send(channel.send_stderr, err)
                 channel.send_exit_status(proc.returncode)
@@ -236,8 +242,8 @@ class SSHServer(paramiko.ServerInterface):
         return False
 
 
-def make_ssh_server_connection(conn, none_auth=False, password_auth=None):
-    log("make_ssh_server_connection%s", (conn, none_auth, password_auth))
+def make_ssh_server_connection(conn, socket_options, none_auth=False, password_auth=None):
+    log("make_ssh_server_connection%s", (conn, socket_options, none_auth, password_auth))
     ssh_server = SSHServer(none_auth=none_auth, password_auth=password_auth)
     DoGSSAPIKeyExchange = False
     sock = conn._socket
@@ -254,51 +260,68 @@ def make_ssh_server_connection(conn, none_auth=False, password_auth=None):
             conn.close()
         except Exception:
             log("%s.close()", conn)
-    ssh_key_dirs = get_ssh_conf_dirs()
-    log("ssh key dirs=%s", ssh_key_dirs)
     try:
         t = paramiko.Transport(sock, gss_kex=DoGSSAPIKeyExchange)
-        t.set_gss_host(socket.getfqdn(""))
+        gss_host = socket_options.get("ssh-gss-host", socket.getfqdn(""))
+        t.set_gss_host(gss_host)
+        #load host keys:
+        PREFIX = "ssh_host_"
+        SUFFIX = "_key"
         host_keys = {}
-        log("trying to load ssh host keys from: %s", csv(ssh_key_dirs))
-        for d in ssh_key_dirs:
-            fd = osexpand(d)
-            log("osexpand(%s)=%s", d, fd)
-            if not os.path.exists(fd) or not os.path.isdir(fd):
-                log("ssh host key directory '%s' is invalid", fd)
-                continue
-            for f in os.listdir(fd):
-                PREFIX = "ssh_host_"
-                SUFFIX = "_key"
-                if f.startswith(PREFIX) and f.endswith(SUFFIX):
-                    ff = os.path.join(fd, f)
-                    keytype = f[len(PREFIX):-len(SUFFIX)]
-                    if keytype:
-                        keyclass = getattr(paramiko, "%sKey" % keytype.upper(), None)
-                        if keyclass is None:
-                            #Ed25519Key
-                            keyclass = getattr(paramiko, "%s%sKey" % (keytype[:1].upper(), keytype[1:]), None)
-                        if keyclass is None:
-                            log("key type %s is not supported, cannot load '%s'", keytype, ff)
-                            continue
-                        log("loading %s key from '%s' using %s", keytype, ff, keyclass)
-                        try:
-                            host_key = keyclass(filename=ff)
-                            if host_key not in host_keys:
-                                host_keys[host_key] = ff
-                                t.add_server_key(host_key)
-                        except IOError as e:
-                            log("cannot add host key '%s'", ff, exc_info=True)
-                        except paramiko.SSHException as e:
-                            log("error adding host key '%s'", ff, exc_info=True)
-                            log.error("Error: cannot add %s host key '%s':", keytype, ff)
-                            log.error(" %s", e)
-        if not host_keys:
-            log.error("Error: cannot start SSH server,")
-            log.error(" no readable SSH host keys found in:")
-            log.error(" %s", csv(ssh_key_dirs))
-            close()
-            return None
+        def add_host_key(fd, f):
+            ff = os.path.join(fd, f)
+            keytype = f[len(PREFIX):-len(SUFFIX)]
+            if not keytype:
+                log.warn("Warning: unknown host key format '%s'", f)
+                return False
+            keyclass = getattr(paramiko, "%sKey" % keytype.upper(), None)
+            if keyclass is None:
+                #Ed25519Key
+                keyclass = getattr(paramiko, "%s%sKey" % (keytype[:1].upper(), keytype[1:]), None)
+            if keyclass is None:
+                log("key type %s is not supported, cannot load '%s'", keytype, ff)
+                return False
+            log("loading %s key from '%s' using %s", keytype, ff, keyclass)
+            try:
+                host_key = keyclass(filename=ff)
+                if host_key not in host_keys:
+                    host_keys[host_key] = ff
+                    t.add_server_key(host_key)
+                    return True
+            except IOError as e:
+                log("cannot add host key '%s'", ff, exc_info=True)
+            except paramiko.SSHException as e:
+                log("error adding host key '%s'", ff, exc_info=True)
+                log.error("Error: cannot add %s host key '%s':", keytype, ff)
+                log.error(" %s", e)
+            return False
+        host_key = socket_options.get("ssh-host-key")
+        if host_key:
+            d, f = os.path.split(host_key)
+            if f.startswith(PREFIX) and f.endswith(SUFFIX):
+                add_host_key(d, f)
+            if not host_keys:
+                log.error("Error: failed to load host key '%s'", host_key)
+                close()
+                return None
+        else:
+            ssh_key_dirs = get_ssh_conf_dirs()
+            log("trying to load ssh host keys from: %s", csv(ssh_key_dirs))
+            for d in ssh_key_dirs:
+                fd = osexpand(d)
+                log("osexpand(%s)=%s", d, fd)
+                if not os.path.exists(fd) or not os.path.isdir(fd):
+                    log("ssh host key directory '%s' is invalid", fd)
+                    continue
+                for f in os.listdir(fd):
+                    if f.startswith(PREFIX) and f.endswith(SUFFIX):
+                        add_host_key(fd, f)
+            if not host_keys:
+                log.error("Error: cannot start SSH server,")
+                log.error(" no readable SSH host keys found in:")
+                log.error(" %s", csv(ssh_key_dirs))
+                close()
+                return None
         log("loaded host keys: %s", tuple(host_keys.values()))
         t.start_server(server=ssh_server)
     except (paramiko.SSHException, EOFError) as e:
@@ -331,4 +354,4 @@ def make_ssh_server_connection(conn, none_auth=False, password_auth=None):
         close()
         return None
     log("client authenticated, channel=%s", chan)
-    return SSHSocketConnection(ssh_server.proxy_channel, sock, conn.local, conn.endpoint, conn.target)
+    return SSHSocketConnection(ssh_server.proxy_channel, sock, conn.local, conn.endpoint, conn.target, socket_options=socket_options)
